# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase, DBTestCase, LogCapture
from anyblok.registry import RegistryManager
from anyblok.config import Configuration
from anyblok.blok import BlokManager, Blok
from anyblok.column import Integer
from threading import Thread
from logging import ERROR


class Test:
    pass


class TestTest:
    pass


class TestTestTest:
    pass


class TestRegistry(DBTestCase):

    def tearDown(self):
        super(TestRegistry, self).tearDown()
        for cls in (Test, TestTest):
            if hasattr(cls, 'Test'):
                delattr(cls, 'Test')

    def test_get_model(self):
        registry = self.init_registry(None)
        System = registry.get('Model.System')
        self.assertEqual(registry.System, System)

    def test_get_the_same_registry(self):
        registry = self.init_registry(None)
        registry2 = self.getRegistry()
        self.assertEqual(registry, registry2)

    def test_reload(self):
        registry = self.init_registry(None)
        bloks_before_reload = registry.System.Blok.query('name').filter(
            registry.System.Blok.state == 'installed').all()
        registry.reload()
        bloks_after_reload = registry.System.Blok.query('name').filter(
            registry.System.Blok.state == 'installed').all()
        self.assertEqual(bloks_before_reload, bloks_after_reload)

    def test_get_bloks_to_load(self):
        registry = self.init_registry(None)
        bloks = registry.get_bloks_to_load()
        self.assertIn('anyblok-core', bloks)

    def test_load_entry(self):
        registry = self.init_registry(None)
        registry.loaded_registries['entry_names'] = []
        RegistryManager.loaded_bloks['blok'] = {
            'entry': {
                'registry_names': ['key'],
                'key': {'properties': {'property': True},
                        'bases': [TestCase]},
            },
        }
        registry.load_entry('blok', 'entry')
        self.assertEqual(registry.loaded_registries['key'],
                         {'properties': {'property': True},
                          'bases': [TestCase]})

    def test_load_core(self):
        RegistryManager.loaded_bloks['blok'] = {
            'Core': {'Session': [TestCase]},
        }
        registry = self.init_registry(None)
        registry.load_core('blok', 'Session')
        self.assertIn(TestCase, registry.loaded_cores['Session'])

    def test_load_blok(self):

        class BlokTest(Blok):
            pass

        registry = self.init_registry(None)
        BlokManager.bloks['blok'] = BlokTest
        RegistryManager.loaded_bloks['blok'] = {
            'Core': {'Session': [TestCase],
                     'Base': [],
                     'SqlBase': [],
                     'SqlViewBase': [],
                     'Query': [],
                     'InstrumentedList': []},
            'removed': [],
            'properties': {},
        }
        for entry in RegistryManager.declared_entries:
            RegistryManager.loaded_bloks['blok'][entry] = {
                'registry_names': []}

        registry.load_blok('blok', False, [])
        self.assertIn(TestCase, registry.loaded_cores['Session'])

    def check_added_in_regisry(self, registry):
        self.assertEqual(registry.Test, Test)
        self.assertEqual(registry.Test.Test, TestTest)
        self.assertEqual(registry.Test.Test.Test, TestTestTest)

    def test_add_in_registry_1(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test', Test)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        self.check_added_in_regisry(registry)

    def test_add_in_registry_2(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test', Test)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        self.check_added_in_regisry(registry)

    def test_add_in_registry_3(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        registry.add_in_registry('Declarations.Test', Test)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        self.check_added_in_regisry(registry)

    def test_add_in_registry_4(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        registry.add_in_registry('Declarations.Test', Test)
        self.check_added_in_regisry(registry)

    def test_add_in_registry_5(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        registry.add_in_registry('Declarations.Test', Test)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        self.check_added_in_regisry(registry)

    def test_add_in_registry_6(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        registry.add_in_registry('Declarations.Test', Test)
        self.check_added_in_regisry(registry)

    def test_registry_db_exist(self):
        self.assertTrue(Configuration.get('Registry').db_exists(
            db_name=Configuration.get('db_name')))

    def test_registry_db_unexist(self):
        self.assertFalse(Configuration.get('Registry').db_exists(
            db_name='wrong_db_name'))


class TestRegistry2(DBTestCase):

    def add_model(self):

        from anyblok import Declarations

        register = Declarations.register
        Model = Declarations.Model

        @register(Model)
        class Test:

            id = Integer(primary_key=True)
            val = Integer(default=0)

            @classmethod
            def _precommit_hook(cls):
                for t in cls.query().all():
                    t.val += t.id

            def add_precommit_hook(self):
                self.precommit_hook('_precommit_hook')

            @classmethod
            def add_cl_precommit_hook(cls):
                cls.precommit_hook('_precommit_hook')

    def test_precommit_hook(self):
        registry = self.init_registry(self.add_model)
        t1 = registry.Test.insert()
        t1.add_precommit_hook()
        t2 = registry.Test.insert()
        t2.add_precommit_hook()
        self.assertEqual(t1.val, 0)
        self.assertEqual(t2.val, 0)
        registry.commit()
        self.assertEqual(t1.val, t1.id)
        self.assertEqual(t2.val, t2.id)
        registry.commit()
        self.assertEqual(t1.val, t1.id)
        self.assertEqual(t2.val, t2.id)
        registry.Test.add_cl_precommit_hook()
        self.assertEqual(t1.val, t1.id)
        self.assertEqual(t2.val, t2.id)
        registry.commit()
        self.assertEqual(t1.val, 2 * t1.id)
        self.assertEqual(t2.val, 2 * t2.id)
        registry.commit()
        self.assertEqual(t1.val, 2 * t1.id)
        self.assertEqual(t2.val, 2 * t2.id)
        t1.add_precommit_hook()
        self.assertEqual(t1.val, 2 * t1.id)
        self.assertEqual(t2.val, 2 * t2.id)
        registry.commit()
        self.assertEqual(t1.val, 3 * t1.id)
        self.assertEqual(t2.val, 3 * t2.id)

    def test_precommit_hook_in_thread(self):
        registry = self.init_registry(self.add_model)
        t1 = registry.Test.insert()
        t1.add_precommit_hook()
        t2 = registry.Test.insert()
        t2.add_precommit_hook()
        self.assertEqual(t1.val, 0)
        self.assertEqual(t2.val, 0)

        def target():
            registry.commit()

        t = Thread(target=target)
        t.start()
        t.join()

        self.assertEqual(t1.val, 0)
        self.assertEqual(t2.val, 0)
        registry.commit()
        self.assertEqual(t1.val, t1.id)
        self.assertEqual(t2.val, t2.id)

    def test_precommit_hook_in_thread_2(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _precommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

        registry = self.init_registry(add_in_registry)
        self.assertEqual(do_somthing, 0)

        def target():
            registry.Test.precommit_hook('_precommit_hook')
            registry.commit()

        t = Thread(target=target)
        t.start()
        t.join()

        registry.commit()
        self.assertEqual(do_somthing, 1)

    def define_cls(self, typename='Model', name='Test', val=1, usesuper=False,
                   inherit=None):

        from anyblok import Declarations

        register = Declarations.register
        Type = getattr(Declarations, typename)
        if inherit is None:
            inherit = object
        else:
            inherit = getattr(Declarations.Mixin, inherit)

        @register(Type, name_=name)
        class Test(inherit):

            @classmethod
            def foo(cls):
                if usesuper:
                    return val * super(Test, cls).foo()

                return val

        return Test

    def test_check_define_cls(self):

        def add_in_registry():
            self.define_cls()

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 1)

    def test_check_define_cls_with_inherit(self):

        def add_in_registry():
            self.define_cls()
            self.define_cls(val=2, usesuper=True)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 2)

    def test_check_define_cls_with_inherit_core(self):

        def add_in_registry():
            self.define_cls(typename='Core', name='Base', val=2)
            self.define_cls(val=2, usesuper=True)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 4)

    def test_check_define_cls_with_inherit_mixin(self):

        def add_in_registry():
            self.define_cls(typename='Mixin', name='MTest', val=3)
            self.define_cls(val=3, usesuper=True, inherit='MTest')

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 9)

    def test_check_define_cls_with_inherit2(self):

        def add_in_registry():
            self.define_cls()
            self.define_cls(val=2, usesuper=True)
            self.define_cls(val=2, usesuper=True)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 4)

    def test_check_define_cls_with_inherit_core2(self):

        def add_in_registry():
            self.define_cls(typename='Core', name='Base', val=2)
            self.define_cls(typename='Core', name='Base', val=2, usesuper=True)
            self.define_cls(val=2, usesuper=True)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 8)

    def test_check_define_cls_with_inherit_mixin2(self):

        def add_in_registry():
            self.define_cls(typename='Mixin', name='MTest', val=3)
            self.define_cls(typename='Mixin', name='MTest', val=3,
                            usesuper=True)
            self.define_cls(val=3, usesuper=True, inherit='MTest')

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 27)

    def test_remove(self):

        def add_in_registry():
            self.define_cls()
            cls_ = self.define_cls(val=2, usesuper=True)
            self.define_cls(val=2, usesuper=True)
            from anyblok import Declarations
            Declarations.unregister(Declarations.Model.Test, cls_)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 2)

    def test_remove_core(self):

        def add_in_registry():
            self.define_cls(typename='Core', name='Base', val=2)
            cls_ = self.define_cls(typename='Core', name='Base', val=2,
                                   usesuper=True)
            self.define_cls(val=2, usesuper=True)
            from anyblok import Declarations
            Declarations.unregister(Declarations.Core.Base, cls_)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 4)

    def test_remove_mixin(self):

        def add_in_registry():
            self.define_cls(typename='Mixin', name='MTest', val=3)
            cls_ = self.define_cls(typename='Mixin', name='MTest', val=3,
                                   usesuper=True)
            self.define_cls(val=3, usesuper=True, inherit='MTest')
            from anyblok import Declarations
            Declarations.unregister(Declarations.Mixin.MTest, cls_)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 9)

    def test_postcommit_hook(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                id = Integer(primary_key=True)

                @classmethod
                def _postcommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

                def add_precommit_hook(self):
                    self.postcommit_hook('_postcommit_hook')

                @classmethod
                def add_cl_precommit_hook(cls):
                    cls.postcommit_hook('_postcommit_hook')

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        t.add_precommit_hook()
        self.assertEqual(do_somthing, 0)
        registry.commit()
        self.assertEqual(do_somthing, 1)
        registry.commit()
        self.assertEqual(do_somthing, 1)
        registry.Test.add_cl_precommit_hook()
        self.assertEqual(do_somthing, 1)
        registry.commit()
        self.assertEqual(do_somthing, 2)
        registry.commit()
        self.assertEqual(do_somthing, 2)
        t.add_precommit_hook()
        self.assertEqual(do_somthing, 2)
        registry.commit()
        self.assertEqual(do_somthing, 3)

    def test_postcommit_hook_in_thread(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _postcommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

        registry = self.init_registry(add_in_registry)
        registry.Test.postcommit_hook('_postcommit_hook')
        self.assertEqual(do_somthing, 0)

        def target():
            registry.commit()

        t = Thread(target=target)
        t.start()
        t.join()

        self.assertEqual(do_somthing, 0)
        registry.commit()
        self.assertEqual(do_somthing, 1)

    def test_postcommit_hook_in_thread_2(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _postcommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

        registry = self.init_registry(add_in_registry)
        self.assertEqual(do_somthing, 0)

        def target():
            registry.Test.postcommit_hook('_postcommit_hook')
            registry.commit()

        t = Thread(target=target)
        t.start()
        t.join()

        self.assertEqual(do_somthing, 1)

    def test_postcommit_hook_call_only_if_commited(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _precommit_hook(cls):
                    raise Exception('Test')

                @classmethod
                def _postcommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

        registry = self.init_registry(add_in_registry)
        registry.Test.postcommit_hook('_postcommit_hook')
        self.assertEqual(do_somthing, 0)
        registry.commit()
        self.assertEqual(do_somthing, 1)
        registry.Test.precommit_hook('_precommit_hook')
        registry.Test.postcommit_hook('_postcommit_hook')
        with self.assertRaises(Exception):
            registry.commit()

        self.assertEqual(do_somthing, 1)

    def test_postcommit_hook_call_only_if_raised(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _precommit_hook(cls):
                    raise Exception('Test')

                @classmethod
                def _postcommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

        registry = self.init_registry(add_in_registry)
        registry.Test.postcommit_hook(
            '_postcommit_hook', call_only_if='raised')
        self.assertEqual(do_somthing, 0)
        registry.commit()
        self.assertEqual(do_somthing, 0)
        registry.Test.precommit_hook('_precommit_hook')
        registry.Test.postcommit_hook(
            '_postcommit_hook', call_only_if='raised')
        with self.assertRaises(Exception):
            registry.commit()

        self.assertEqual(do_somthing, 1)

    def test_postcommit_hook_call_only_if_always(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _precommit_hook(cls):
                    raise Exception('Test')

                @classmethod
                def _postcommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

        registry = self.init_registry(add_in_registry)
        registry.Test.postcommit_hook(
            '_postcommit_hook', call_only_if='always')
        self.assertEqual(do_somthing, 0)
        registry.commit()
        self.assertEqual(do_somthing, 1)
        registry.Test.precommit_hook('_precommit_hook')
        registry.Test.postcommit_hook(
            '_postcommit_hook', call_only_if='always')
        with self.assertRaises(Exception):
            registry.commit()

        self.assertEqual(do_somthing, 2)

    def test_postcommit_hook_with_rollback(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _postcommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

        registry = self.init_registry(add_in_registry)
        registry.Test.postcommit_hook(
            '_postcommit_hook')
        self.assertEqual(do_somthing, 0)
        registry.rollback()
        registry.commit()
        self.assertEqual(do_somthing, 0)

    def test_postcommit_hook_with_exception(self):

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _postcommit_hook(cls):
                    raise Exception('Here one exception')

        registry = self.init_registry(add_in_registry)
        registry.Test.postcommit_hook(
            '_postcommit_hook')
        with LogCapture('anyblok.registry', level=ERROR) as logs:
            registry.commit()
            messages = logs.get_error_messages()
            message = messages[0]
            self.assertIn('Here one exception', message)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
import datetime
from json import loads, dumps
from .exceptions import FormaterException


register = Declarations.register
IO = Declarations.Model.IO


@register(IO)
class Formater:

    def str2value(self, value, model):
        return value

    def _externalIdStr2value(self, value, model):
        mapping = self.registry.IO.Mapping.get(model, value)
        if mapping is None:
            raise FormaterException(
                "Unexisting maping key %r with model %r" % (value, model))

        return mapping

    def externalIdStr2value(self, value, model):
        entry = self._externalIdStr2value(value, model)
        pks = entry.to_primary_keys()
        if len(pks.keys()) > 1:
            raise FormaterException(
                "Foreign key on multi primary keys does not implemented yet")

        return [x for x in pks.values()][0]

    def value2str(self, value, model):
        if value is None:
            return ''

        return str(value)

    def externalIdValue2str(self, value, model):
        Model = self.registry.get(model)
        Mapping = self.registry.IO.Mapping
        pks = Model.get_primary_keys()
        if len(pks) > 1:
            raise FormaterException(
                "Foreign key on multi primary keys does not implemented yet")

        pks = {x: value for x in pks}
        mapping = Mapping.get_from_model_and_primary_keys(model, pks)

        if mapping is None:
            entry = Model.from_primary_keys(**pks)
            key = self.registry.IO.Exporter.get_external_id(model)
            Mapping.set(key, entry)
            return key

        return mapping.key


@register(IO.Formater)
class Float(IO.Formater):

    def str2value(self, value, model):
        return float(value)


@register(IO.Formater)
class Decimal(IO.Formater):

    def str2value(self, value, model):
        from decimal import Decimal as D
        return D(value)


@register(IO.Formater)
class Json(IO.Formater):

    def str2value(self, value, model):
        return loads(value)

    def value2str(self, value, model):
        return dumps(value)


@register(IO.Formater)
class Interval(IO.Formater):

    def str2value(self, value, model):
        return datetime.timedelta(seconds=int(value))

    def value2str(self, value, model):
        return str(value.seconds)


@register(IO.Formater)
class Integer(IO.Formater):

    def str2value(self, value, model):
        return int(value)


@register(IO.Formater)
class SmallInteger(IO.Formater.Integer):
    pass


@register(IO.Formater)
class BigInteger(IO.Formater.Integer):
    pass


@register(IO.Formater)
class Boolean(IO.Formater):

    def str2value(self, value, model):
        if value in ("1", "true", "True", 1, True):
            return True
        elif value in ("0", "false", "False", '', 0, False):
            return False

        raise FormaterException(
            "Value %r is not a boolean" % value)

    def value2str(self, value, model):
        return '1' if value else '0'


@register(IO.Formater)
class Time(IO.Formater):

    def str2value(self, value, model):
        dt = datetime.datetime.strptime(value, "%H:%M:%S")
        return datetime.time(dt.hour, dt.minute, dt.second)


@register(IO.Formater)
class Date(IO.Formater):

    def str2value(self, value, model):
        dt = datetime.datetime.strptime(value, "%Y-%m-%d")
        return datetime.date(dt.year, dt.month, dt.day)


@register(IO.Formater)
class DateTime(IO.Formater):

    def str2value(self, value, model):
        return datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")


@register(IO.Formater)
class Many2One(IO.Formater):

    def str2value(self, value, model):
        Model = self.registry.get(model)
        if not value:
            return None

        pks = loads(value)
        if not pks:
            return None

        if not isinstance(pks, dict):
            raise FormaterException(
                "Value %r for %r must be dict" % (
                    value, self.__registry_name__))

        return Model.from_primary_keys(**pks)

    def externalIdStr2value(self, value, model):
        return self._externalIdStr2value(value, model)

    def value2str(self, value, model):
        if value is None:
            return ''

        return dumps(value.to_primary_keys())

    def externalIdValue2str(self, value, model):
        if value is None:
            return ''

        Exporter = self.registry.IO.Exporter
        return Exporter.get_key_mapping(value)


@register(IO.Formater)
class One2One(IO.Formater.Many2One):
    pass


@register(IO.Formater)
class Many2Many(IO.Formater):

    def str2value(self, value, model):
        Model = self.registry.get(model)
        if not value:
            return None

        pks = loads(value)

        if not all(isinstance(x, dict) for x in pks):
            raise FormaterException(
                "All values in %r for %r must be dict" % (
                    value, self.__registry_name__))

        return [Model.from_primary_keys(**x) for x in pks if x]

    def externalIdStr2value(self, values, model):
        if not values:
            return None
        values = loads(values)
        return [self._externalIdStr2value(value, model) for value in values]

    def value2str(self, values, model):
        if not values:
            return dumps([])

        return dumps([value.to_primary_keys() for value in values])

    def externalIdValue2str(self, values, model):
        if not values:
            return dumps([])

        Exporter = self.registry.IO.Exporter
        return dumps([Exporter.get_key_mapping(value) for value in values])


@register(IO.Formater)
class One2Many(IO.Formater.Many2Many):
    pass



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from nose.plugins import Plugin
from anyblok.config import Configuration, get_db_name
from anyblok import (
    load_init_function_from_entry_points,
    configuration_post_load,
)
from anyblok.blok import BlokManager
from anyblok.registry import RegistryManager, return_list
from os.path import join
from os import walk


class Arg2OptOptions:

    def __init__(self, options):
        self.options = options

    def _get_args(self):
        return False

    def _get_kwargs(self):
        keys = ['configfile', 'db_name', 'db_driver_name', 'db_user_name',
                'db_password', 'db_host', 'db_port']

        return [(x, getattr(self.options, x))
                for x in keys] + [('withoutautomigration', True)]

    def __getattr__(self, name, default=None):
        return getattr(self.options, name, default)


class AnyBlokPlugin(Plugin):
    name = 'anyblok-bloks'
    score = 100

    def __init__(self):
        super(AnyBlokPlugin, self).__init__()
        self.authoried_bloks_test_files = []
        self.bloks_path = []
        self.registryLoaded = False
        self.AnyBlokOptions = None

    def options(self, parser, env):
        super(AnyBlokPlugin, self).options(parser, env)
        parser.add_option("--anyblok-configfile", dest="configfile")
        parser.add_option('--anyblok-db-name', dest='db_name',
                          default=env.get('ANYBLOK_DATABASE_NAME'),
                          help="Name of the database")
        parser.add_option('--anyblok-db-driver-name', dest='db_driver_name',
                          default=env.get('ANYBLOK_DATABASE_DRIVER'),
                          help="the name of the database backend. This name "
                               "will correspond to a module in "
                               "sqlalchemy/databases or a third party plug-in")
        parser.add_option('--anyblok-db-user-name', dest='db_user_name',
                          default=env.get('ANYBLOK_DATABASE_USER'),
                          help="The user name")
        parser.add_option('--anyblok-db-password', dest='db_password',
                          default=env.get('ANYBLOK_DATABASE_PASSWORD'),
                          help="database password")
        parser.add_option('--anyblok-db-host', dest='db_host',
                          default=env.get('ANYBLOK_DATABASE_HOST'),
                          help="The name of the host")
        parser.add_option('--anyblok-db-port', dest='db_port',
                          default=env.get('ANYBLOK_DATABASE_PORT'),
                          help="The port number")
        parser.add_option('--anyblok-db-url', dest='db_url',
                          default=env.get('ANYBLOK_DATABASE_URL'),
                          help="Complete URL for connection with the database")

    def configure(self, options, conf):
        super(AnyBlokPlugin, self).configure(options, conf)
        if self.enabled:
            self.AnyBlokOptions = Arg2OptOptions(options)

    def load_registry(self):
        if self.enabled and self.registryLoaded is False:
            # Load the registry here not in configuration,
            # because the configuration are not load in order of score
            self.registryLoaded = True
            load_init_function_from_entry_points(unittest=True)
            Configuration.load_config_for_test()
            Configuration.parse_options(self.AnyBlokOptions)
            configuration_post_load()
            BlokManager.load()
            db_name = get_db_name()

            registry = RegistryManager.get(db_name)
            if registry:
                installed_bloks = registry.System.Blok.list_by_state(
                    "installed")
                selected_bloks = return_list(
                    Configuration.get('selected_bloks')) or installed_bloks

                unwanted_bloks = return_list(
                    Configuration.get('unwanted_bloks')) or []

                self.bloks_path = [BlokManager.getPath(x)
                                   for x in BlokManager.ordered_bloks]

                self.authoried_bloks_test_files = []
                for blok in installed_bloks:
                    if blok not in selected_bloks or blok in unwanted_bloks:
                        continue

                    startpath = BlokManager.getPath(blok)
                    for root, dirs, _ in walk(startpath):
                        if 'tests' in dirs:
                            self.authoried_bloks_test_files.append(
                                join(root, 'tests'))

                registry.close()  # free the registry to force create it again

    def file_from_blok(self, file):
        for blok_path in self.bloks_path:
            if file.startswith(blok_path):
                return True

        return False

    def file_from_authorized_bloks(self, file):
        for testFile in self.authoried_bloks_test_files:
            if file.startswith(testFile):
                return True

        return False

    def wantModule(self, module):
        self.load_registry()
        return True

    def wantFile(self, file, package=None):
        self.load_registry()
        if self.enabled:
            if file.endswith(".py"):
                if self.file_from_blok(file):
                    return self.file_from_authorized_bloks(file)

        return None



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from ..exceptions import CoreBaseException
register = Declarations.register


@register(Declarations.Core)
class Base:
    """ Inherited by all the models
    """

    is_sql = False

    @classmethod
    def initialize_model(cls):
        """ This method is called to initialize a model during the creation of
        the registry
        """
        pass

    @classmethod
    def fire(cls, event, *args, **kwargs):
        """ Call a specific event on the model

        :param event: Name of the event
        """
        events = cls.registry.events
        if cls.__registry_name__ in events:
            if event in events[cls.__registry_name__]:
                for model, method in events[cls.__registry_name__][event]:
                    m = cls.registry.get(model)
                    getattr(m, method)(*args, **kwargs)

    @classmethod
    def get_model(cls, model):
        return cls.registry.get(model)

    @classmethod
    def get_primary_keys(cls, **pks):
        """ No SQL Model has not primary key """
        raise CoreBaseException("No primary key for No SQL Model")

    @classmethod
    def from_primary_keys(cls, **pks):
        """ No SQL Model has not primary key """
        raise CoreBaseException("No primary key for No SQL Model")

    def to_primary_keys(self):
        """ No SQL Model has not primary key """
        raise CoreBaseException("No primary key for No SQL Model")

    def has_perm(self, principals, permission):
        """Check that one of principals has permission on given record.

        Since this is an ordinary instance method, it can't be used on the
        model class itself. For this use case, see :meth:`has_model_perm`
        """
        return self.registry.check_permission(self, principals, permission)

    @classmethod
    def has_model_perm(cls, principals, permission):
        """Check that one of principals has permission on given model.

        Since this is a classmethod, even if called on a record, only its
        model class will be considered for the permission check.
        """
        return cls.registry.check_permission(cls, principals, permission)

    @classmethod
    def precommit_hook(cls, method, *args, **kwargs):
        """ Same in the registry a hook to call just before the commit

        .. warning::

            Only one instance with same parameters of the hook is called before
            the commit

        :param method: the method to call on this model
        :param put_at_the_end_if_exist: If ``True`` the hook is move at the end
        """
        cls.registry.precommit_hook(
            cls.__registry_name__, method, *args, **kwargs)

    @classmethod
    def postcommit_hook(cls, method, *args, **kwargs):
        """ Same in the registry a hook to call just after the commit

        you can choice if the hook is called in function of ``call_only_if``:

        * ``commited``: Call if the commit is done without exception
        * ``raised``: Call if one exception was raised
        * ``always``: Always call

        .. warning::

            Only one instance with same paramters of the hook is called
            after the commit

        :param method: the method to call on this model
        :param put_at_the_end_if_exist: If ``True`` the hook is move at the end
        :param call_only_if: ['commited' (default), 'raised', 'always']
        """
        cls.registry.postcommit_hook(
            cls.__registry_name__, method, *args, **kwargs)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from sqlalchemy.ext.hybrid import hybrid_property
from anyblok.common import anyblok_column_prefix


class FieldException(Exception):
    """ Simple Exception for Field """


class Field:
    """ Field class

    This class must not be instanciated
    """

    use_hybrid_property = False

    def __init__(self, *args, **kwargs):
        """ Initialize the field

        :param label: label of this field
        :type label: str
        """
        self.forbid_instance(Field)
        self.label = None

        if 'label' in kwargs:
            self.label = kwargs.pop('label')

        self.context = kwargs.pop('context', {})
        self.args = args
        self.kwargs = kwargs

    def forbid_instance(self, cls):
        """ Raise an exception if the cls is an instance of this __class__

        :param cls: instance of the class
        :exception: FieldException
        """
        if self.__class__ is cls:
            raise FieldException(
                "%r class must not be instanciated use a sub class" % cls)

    def update_description(self, registry, model, res):
        res.update(self.context)

    def update_properties(self, registry, namespace, fieldname, properties):
        """ Update the propertie use to add new column

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: properties known to the model
        """

    def get_property(self, registry, namespace, fieldname, properties):
        """Return the property of the field

        .. warning::

            In the case of the get is called in classattribute,
            SQLAlchemy wrap for each call the column, the id of the wrapper
            is not the same

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: properties known to the model
        """
        return hybrid_property(
            self.wrap_getter_column(fieldname),
            self.wrap_setter_column(fieldname),
            expr=self.wrap_expr_column(fieldname))

    def getter_format_value(self, value):
        return value

    def wrap_getter_column(self, fieldname):
        """Return a default getter for the field

        :param fieldname: name of the field
        """
        attr_name = anyblok_column_prefix + fieldname

        def getter_column(model_self):
            return self.getter_format_value(getattr(model_self, attr_name))

        return getter_column

    def wrap_expr_column(self, fieldname):
        """Return a default expr for the field

        :param fieldname: name of the field
        """
        attr_name = anyblok_column_prefix + fieldname

        def expr_column(model_self):
            return getattr(model_self, attr_name)

        return expr_column

    def expire_related_attribute(self, model_self, action_todos):
        for action_todo in action_todos:
            if len(action_todo) == 1:
                obj = model_self
                attrs = [action_todo[0]]
            else:
                obj = getattr(model_self, action_todo[0])
                attrs = [action_todo[1]]
                if obj is None:
                    continue

            if obj in model_self.registry.session:
                if obj._sa_instance_state.persistent:
                    model_self.registry.expire(obj, attrs)

    def setter_format_value(self, value):
        return value

    def wrap_setter_column(self, fieldname):
        attr_name = anyblok_column_prefix + fieldname

        def setter_column(model_self, value):
            action_todos = set()
            if fieldname in model_self.loaded_columns:
                action_todos = model_self.registry.expire_attributes.get(
                    model_self.__registry_name__, {}).get(fieldname, set())

            self.expire_related_attribute(model_self, action_todos)
            value = self.setter_format_value(value)
            res = setattr(model_self, attr_name, value)
            self.expire_related_attribute(model_self, action_todos)
            return res

        return setter_column

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """ Return the instance of the real field

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: properties known of the model
        :rtype: instance of Field
        """
        self.format_label(fieldname)
        return self

    def format_label(self, fieldname):
        """ Return the label for this field

        :param fieldname: if no label filled, the fieldname will be capitalized
            and returned
        :rtype: the label for this field
        """
        if not self.label:
            label = fieldname.replace('_', ' ')
            self.label = label.capitalize()

    def native_type(self):
        """ Return the native SqlAlchemy type

        :exception: FieldException
        """
        raise FieldException("No native type for this field")

    def must_be_declared_as_attr(self):
        """ Return False, it is the default value """
        return False

    def must_be_duplicate_before_added(self):
        """ Return False, it is the default value """
        return False

    def autodoc_get_properties(self):
        res = {'Field type': str(self.__class__),
               'Context': self.context,
               'Label': self.label}
        res.update(self.kwargs)
        return res

    def autodoc_format_dict(self, key, value, padding=0):
        key = key.strip()
        if isinstance(value, dict):
            res = ' * ``%s``:\n' % key
            res += '\n'.join(
                [self.autodoc_format_dict(x, y, padding=padding + 1)
                 for x, y in value.items()])
            res += '\n'
            return res
        elif isinstance(value, (list, tuple)):
            res = ' * ``%s``:\n' % key
            res += '\n'.join([' ' * padding + ' * %s' % str(x)
                              for x in value])
            res += '\n'
            return res
        else:
            return ' ' * padding + ' * ``%s`` - %s' % (key, str(value))

    def autodoc(self):
        return '\n'.join([self.autodoc_format_dict(x, y)
                          for x, y in self.autodoc_get_properties().items()])


class Function(Field):
    """ Function Field

    ::

        from anyblok.declarations import Declarations
        from anyblok.field import Function


        @Declarations.register(Declarations.Model)
        class Test:
            x = Function(fget='fget', fset='fset', fdel='fdel', fexp='fexpr')

        ..warning::

            fexp must be a classmethod

    """

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """Return the property of the field

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: properties known to the model
        """

        def wrap(method):
            m = self.kwargs.get(method)
            if m is None:
                return None

            def function_method(model_self, *args, **kwargs):
                try:
                    return getattr(model_self, m)(*args, **kwargs)
                except TypeError:
                    if method == 'fget':
                        raise FieldException("You must declare 'fexp' for "
                                             "'%s: %s' field" % (namespace,
                                                                 fieldname))
                    else:
                        raise

            return function_method

        fget = wrap('fget')
        fset = wrap('fset')
        fdel = wrap('fdel')
        fexpr = wrap('fexpr')

        self.format_label(fieldname)
        properties['loaded_fields'][fieldname] = self.label
        return hybrid_property(fget, fset, fdel=fdel, expr=fexpr)



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import os
import sys
from setuptools import setup, find_packages
version = '0.17.1'


if sys.version_info < (3, 3):
    sys.stderr.write("This package requires Python 3.3 or newer. "
                     "Yours is " + sys.version + os.linesep)
    sys.exit(1)

requires = [
    'sqlalchemy',
    'sqlalchemy-utils >= 0.33.0',
    'argparse',
    'alembic',
    'graphviz',
    'nose',  # for unittest during the blok install
    'lxml',
    'six',
    'PyYAML',
    'appdirs',
    'sqlalchemy-utils',
    'pytz',
    'python-dateutil',
    'texttable',
    'testfixtures',
]

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst'), 'r', encoding='utf-8') as readme:
    README = readme.read()

with open(
    os.path.join(here, 'doc', 'CHANGES.rst'), 'r', encoding='utf-8'
) as change:
    CHANGE = change.read()

with open(
    os.path.join(here, 'doc', 'FRONT.rst'), 'r', encoding='utf-8'
) as front:
    FRONT = front.read()

setup(
    name="AnyBlok",
    version=version,
    author="Jean-Sébastien Suzanne",
    author_email="jssuzanne@anybox.fr",
    description="Anyblok is a dynamic injection blok framework",
    license="MPL2",
    long_description=README + '\n' + FRONT + '\n' + CHANGE,
    url="http://docs.anyblok.org/%s" % version,
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    install_requires=requires,
    tests_require=requires + ['nose'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
    ],
    entry_points={
        'console_scripts': [
            'anyblok_createdb=anyblok.scripts:anyblok_createdb',
            'anyblok_updatedb=anyblok.scripts:anyblok_updatedb',
            'anyblok_nose=anyblok.scripts:anyblok_nose',
            'anyblok_interpreter=anyblok.scripts:anyblok_interpreter',
            'anyblok_doc=anyblok.scripts:anyblok2doc',
        ],
        'bloks': [
            'anyblok-core=anyblok.bloks.anyblok_core:AnyBlokCore',
            'model_authz='
            'anyblok.bloks.model_authz:ModelBasedAuthorizationBlok',
            'anyblok-io=anyblok.bloks.io:AnyBlokIO',
            'anyblok-io-csv=anyblok.bloks.io_csv:AnyBlokIOCSV',
            'anyblok-io-xml=anyblok.bloks.io_xml:AnyBlokIOXML',
        ],
        'test_bloks': [
            'test-blok1=anyblok.test_bloks.test_blok1:TestBlok',
            'test-blok2=anyblok.test_bloks.test_blok2:TestBlok',
            'test-blok3=anyblok.test_bloks.test_blok3:TestBlok',
            'test-blok4=anyblok.test_bloks.test_blok4:TestBlok',
            'test-blok5=anyblok.test_bloks.test_blok5:TestBlok',
            'test-blok6=anyblok.test_bloks.test_blok6:TestBlok',
            'test-blok7=anyblok.test_bloks.test_blok7:TestBlok',
            'test-blok8=anyblok.test_bloks.test_blok8:TestBlok',
            'test-blok9=anyblok.test_bloks.test_blok9:TestBlok',
            'test-blok10=anyblok.test_bloks.test_blok10:TestBlok',
            'test-blok11=anyblok.test_bloks.test_blok11:TestBlok',
            'test-blok12=anyblok.test_bloks.test_blok12:TestBlok',
            'test-blok13=anyblok.test_bloks.test_blok13:TestBlok',
        ],
        'nose.plugins.0.10': [
            'anyblok-bloks=anyblok.plugins:AnyBlokPlugin',
        ],
        'anyblok.init': [],
        'anyblok_configuration.post_load': [],
        'anyblok.model.plugin': [
            'hybrid_method=anyblok.model.hybrid_method:HybridMethodPlugin',
            'table_mapper=anyblok.model.table_and_mapper:TableMapperPlugin',
            'event=anyblok.model.event:EventPlugin',
            'sqla-event=anyblok.model.event:SQLAlchemyEventPlugin',
            'auto-orm-event=anyblok.model.event:AutoSQLAlchemyORMEventPlugin',
            'cache=anyblok.model.cache:CachePlugin',
            'field_datetime=anyblok.model.field_datetime:AutoUpdatePlugin',
        ],
    },
    extras_require={},
)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase
from anyblok import Declarations
from anyblok.declarations import DeclarationsException  # FIXME USE Declaration


class OneType:

    @classmethod
    def register(cls, parent, name, cls_, **kwargs):
        setattr(parent, name, cls_)

    @classmethod
    def unregister(cls, entry, cls_):
        pass


class TestDeclaration(TestCase):

    def tearDown(self):
        super(TestDeclaration, self).tearDown()
        for type_ in ['OneType', 'MyOneType']:
            if type_ in Declarations.declaration_types:
                del Declarations.declaration_types[type_]

    def test_add(self):
        Declarations.add_declaration_type(cls_=OneType)
        self.assertEqual(Declarations.declaration_types['OneType'], OneType)

        class SubType:
            pass

        Declarations.register(Declarations.OneType, cls_=SubType)
        Declarations.unregister(Declarations.OneType, cls_=SubType)

    def test_add_doublon(self):
        Declarations.add_declaration_type(cls_=OneType)
        try:
            Declarations.add_declaration_type(cls_=OneType)
            self.fail('No watch dog for doublon declarations type')
        except DeclarationsException:
            pass

    def test_add_decoration(self):

        @Declarations.add_declaration_type()
        class MyOneType(OneType):
            pass

        self.assertEqual(Declarations.declaration_types['MyOneType'],
                         MyOneType)

        @Declarations.register(Declarations.MyOneType)
        class SubType:
            pass

        Declarations.unregister(Declarations.MyOneType.SubType, SubType)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.blok import BlokManager


@Declarations.register(Declarations.Model.Documentation)
class Blok:

    def __init__(self, blok):
        self.blok = blok

    def exist(self):
        return BlokManager.has(self.blok.name)

    @classmethod
    def filterBloks(cls, query):
        Blok = cls.registry.System.Blok
        return query.filter(Blok.state == 'installed').order_by(Blok.order)

    @classmethod
    def getelements(cls):
        Blok = cls.registry.System.Blok
        query = Blok.query()
        query = cls.filterBloks(query)
        return query.all()

    @classmethod
    def header2RST(cls, doc):
        doc.write("Bloks\n=====\n\n")

    @classmethod
    def footer2RST(cls, doc):

        pass

    def toRST(self, doc):
        doc.write(self.blok.name + '\n' + '-' * len(self.blok.name) + '\n\n')
        doc.write(self.blok.short_description + '\n\n')
        self.toRST_write_params(doc)
        doc.write(self.blok.long_description + '\n\n')

    def toRST_get_field(self):
        return ('author', 'version', 'installed_version')

    def toRST_write_params(self, doc):
        fields = self.toRST_get_field()
        msg = "Parameter:\n\n* "
        msg += '\n* '.join('**%s** = %s' % (f, getattr(self.blok, f))
                           for f in fields)
        doc.write(msg + '\n\n')



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase


class TestImporter(BlokTestCase):

    def create_importer(self, **kwargs):
        Importer = self.registry.IO.Importer
        kwargs['model'] = 'Model.IO.Importer'
        if 'check_import' not in kwargs:
            kwargs['check_import'] = False

        if 'commit_at_each_grouped' not in kwargs:
            kwargs['commit_at_each_grouped'] = True

        return Importer(**kwargs)

    def test_commit_if_check(self):
        importer = self.create_importer(check_import=True)
        self.assertEqual(importer.commit(), False)

    def test_commit_if_not_commit_at_each_group(self):
        importer = self.create_importer(commit_at_each_grouped=False)
        self.assertEqual(importer.commit(), False)

    def test_commit(self):
        importer = self.create_importer()
        self.assertEqual(importer.commit(), True)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase, DBTestCase
from anyblok.field import Field, FieldException, Function
from anyblok.column import Integer, String
from anyblok import Declarations
from sqlalchemy import func


Model = Declarations.Model
register = Declarations.register


class OneField(Field):
    pass


class TestField(TestCase):

    def test_forbid_instance(self):
        try:
            Field()
            self.fail("Field mustn't be instanciated")
        except FieldException:
            pass

    def test_without_label(self):
        field = OneField()
        field.get_sqlalchemy_mapping(None, None, 'a_field', None)
        self.assertEqual(field.label, 'A field')


def field_without_name():

    @register(Model)
    class Test:

        id = Integer(primary_key=True)
        field = OneField()


class TestField2(DBTestCase):

    def test_field_without_name(self):
        self.init_registry(field_without_name)

    def define_field_function(self):

        @register(Model)
        class Test:

            id = Integer(primary_key=True)
            first_name = String()
            last_name = String()
            name = Function(fget='fget', fset='fset', fdel='fdel',
                            fexpr='fexpr')

            def fget(self):
                return '{0} {1}'.format(self.first_name, self.last_name)

            def fset(self, value):
                self.first_name, self.last_name = value.split(' ', 1)

            def fdel(self):
                self.first_name = self.last_name = None

            @classmethod
            def fexpr(cls):
                return func.concat(cls.first_name, ' ', cls.last_name)

    def test_field_function_fget(self):
        registry = self.init_registry(self.define_field_function)
        t = registry.Test.insert(first_name='Jean-Sebastien',
                                 last_name='SUZANNE')
        self.assertEqual(t.name, 'Jean-Sebastien SUZANNE')
        t = registry.Test.query().first()
        self.assertEqual(t.name, 'Jean-Sebastien SUZANNE')

    def test_field_function_fset(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                _name = String()
                name = Function(fget='fget', fset='fset', fdel='fdel',
                                fexpr='fexpr')

                def fget(self):
                    return self._name

                def fset(self, value):
                    self._name = value

                def fdel(self):
                    self._name = None

                @classmethod
                def fexpr(cls):
                    return cls._name

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert(name='Jean-Sebastien SUZANNE')
        self.assertEqual(t._name, 'Jean-Sebastien SUZANNE')
        t = registry.Test.query().first()
        t.name = 'Mister ANYBLOK'
        self.assertEqual(t._name, 'Mister ANYBLOK')
        t.update(name='Jean-Sebastien SUZANNE')
        self.assertEqual(t._name, 'Jean-Sebastien SUZANNE')

    def test_field_function_fdel(self):
        registry = self.init_registry(self.define_field_function)
        t = registry.Test.insert(first_name='jean-sebastien',
                                 last_name='suzanne')
        del t.name
        self.assertEqual(t.first_name, None)
        self.assertEqual(t.last_name, None)

    def test_field_function_fexpr(self):
        registry = self.init_registry(self.define_field_function)
        registry.Test.insert(first_name='Jean-Sebastien',
                             last_name='SUZANNE')
        t = registry.Test.query().filter(
            registry.Test.name == 'Jean-Sebastien SUZANNE').first()
        self.assertEqual(t.name, 'Jean-Sebastien SUZANNE')

    def test_field_function_without_fexpr(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val1 = Integer()
                val2 = Function(fget='fget', fset='fset')

                def fget(self):
                    return 2 * self.val1

                def fset(self, val):
                    self.val1 = val / 2

        registry = self.init_registry(add_in_registry)
        registry.Test.insert(val1=1)
        with self.assertRaises(FieldException):
            registry.Test.query().filter(
                registry.Test.val2 == 2).first()



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import String, Boolean


register = Declarations.register
System = Declarations.Model.System
Mixin = Declarations.Mixin


@register(System)
class RelationShip(System.Field):

    name = String(primary_key=True)
    model = String(primary_key=True)
    local_column = String()
    remote_column = String()
    remote_name = String()
    remote_model = String(nullable=False)
    remote = Boolean(default=False)
    nullable = Boolean()

    def _description(self):
        res = super(RelationShip, self)._description()
        remote_name = self.remote_name or ''
        res.update(
            nullable=self.nullable,
            model=self.remote_model,
            remote_name=remote_name
        )
        return res

    @classmethod
    def add_field(cls, rname, relation, model, table, ftype):
        """ Insert a relationship definition

        :param rname: name of the relationship
        :param relation: instance of the relationship
        :param model: namespace of the model
        :param table: name of the table of the model
        :param ftype: type of the AnyBlok Field
        """
        local_column = relation.info.get('local_column')
        remote_column = relation.info.get('remote_column')
        remote_model = relation.info.get('remote_model')
        remote_name = relation.info.get('remote_name')
        label = relation.info.get('label')
        nullable = relation.info.get('nullable', True)

        vals = dict(code=table + '.' + rname,
                    model=model, name=rname, local_column=local_column,
                    remote_model=remote_model, remote_name=remote_name,
                    remote_column=remote_column, label=label,
                    nullable=nullable, ftype=ftype)
        cls.insert(**vals)

        if remote_name:
            remote_type = "Many2One"
            if ftype == "Many2One":
                remote_type = "One2Many"
            elif ftype == 'Many2Many':
                remote_type = "Many2Many"
            elif ftype == "One2One":
                remote_type = "One2One"

            m = cls.registry.get(remote_model)
            vals = dict(code=m.__tablename__ + '.' + remote_name,
                        model=remote_model, name=remote_name,
                        local_column=remote_column, remote_model=model,
                        remote_name=rname,
                        remote_column=local_column,
                        label=remote_name.capitalize().replace('_', ' '),
                        nullable=True, ftype=remote_type, remote=True)
            cls.insert(**vals)

    @classmethod
    def alter_field(cls, field, label, ftype):
        pass



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import Integer, Selection, String


@Declarations.register(Declarations.Mixin)
class IOMixin:

    id = Integer(primary_key=True)
    mode = Selection(selections="get_mode_choices", nullable=False)
    model = String(foreign_key=Declarations.Model.System.Model.use('name'),
                   nullable=False)

    @classmethod
    def get_mode_choices(cls):
        return {}

    def get_formater(self, ctype):
        formater_name = 'Model.IO.Formater.' + ctype
        if self.registry.has(formater_name):
            return self.registry.get(formater_name)()
        else:
            return self.registry.IO.Formater()
